unit BigString; {Big Strings}
{Copyright (C) 1994, Jon Shemitz <jon@armory.com> }

interface

type
  StringTypes = (_int, _wrd, _lng, _ptr);
  BigStringHeader = object
                    StringType: StringTypes;
                    DataSize:   byte; {2 or 4}
                    ObjSize:    word; {Save a recalc -
                                       and force dword alignment of tail}
                    end;

  BigStringPtr = ^ BigStrings;
  BigStringPtrPtr = ^ BigStringPtr;
  BigStrings =
    object (BigStringHeader)
    Data: record
          case StringTypes of
            _int: (Int: array[0..$FFF0 div SizeOf(integer)] of integer);
            _wrd: (Wrd: array[0..$FFF0 div SizeOf(word)]    of word);
            _lng: (Lng: array[0..$FFF0 div SizeOf(LongInt)] of LongInt);
            _ptr: (Ptr: array[0..$FFF0 div SizeOf(pointer)] of pointer);
            end;
          {In all cases, Data[0] is the length}

    function GetLength: word;
    function Sort: BigStringPtr; {Returns an index vector}
    function Copy: BigStringPtr; {Returns a COPY of Self}
    function EqualTo(Other: BigStringPtr): boolean;
    end;

function New_BigString(DataType: StringTypes; DataLength: word): BigStringPtr;

procedure Dispose_BigString(BSP: BigStringPtr);

function GenerateIndexVector(Length: word): BigStringPtr;

implementation

function New_BigString(DataType: StringTypes; DataLength: word): BigStringPtr;
var
  Size:   LongInt;
  Result: BigStringPtr;
const
  DataSizes: array[StringTypes] of byte = (2, 2, 4, 4);
begin
  Size := DataSizes[DataType] * LongInt(DataLength+1) + SizeOf(BigStringHeader);
  if Size > $FFF0 then RunError(201) {else Proceed};

  GetMem(Result, Size);
  with Result^ do
    begin
    StringType  := DataType;
    DataSize    := DataSizes[DataType];
    ObjSize     := Size;
    Data.Lng[0] := DataLength; {Nothing too horrible about writing an extra 0}
    end;                       {in Wrd/Int[1]!}
  New_BigString := Result;
end;

procedure Dispose_BigString(BSP: BigStringPtr);
begin
  FreeMem(BSP, BSP^.ObjSize);
end;

function GenerateIndexVector(Length: word): BigStringPtr;
var
  Idx:    word;
  Result: BigStringPtr;
begin
  Result := New_BigString(_Wrd, Length);
  with Result^ do
    for Idx := 1 to Length do
      Data.Wrd[Idx] := Idx;
  GenerateIndexVector := Result;
end;


function BigStrings.GetLength: word;
begin
  case DataSize of
    2: GetLength := Data.Wrd[0];
    4: GetLength := Data.Lng[0];
    {$ifopt R+} else RunError(201); {$endif}
    end;
end;

  type Comparer = function(A, B: word; Data: BigStringPtr): boolean;
       {Is Data[A] < Data[B]?}

  procedure QuickSort( Lo, Hi:      word;
                       A_LT_B:      Comparer;
                       Self, Index: BigStringPtr);
  var
    I, J, Mid, Xchg: word;
  begin
    I := Lo;  J := Hi; Mid := (Lo + Hi) div 2;
    with Index^ do
      repeat
        while A_LT_B(Data.Wrd[I], Data.Wrd[Mid], Self) do Inc(I);
        { Find first value >= midpoint }
        while A_LT_B(Data.Wrd[Mid], Data.Wrd[J], Self) do Dec(J);
        { Find last value <= midpoint  }
        if I <= J then
          begin
          if I < J then begin {XCHG Ith and Jth (index) record}
                        Xchg := Data.Wrd[I];
                        Data.Wrd[I] := Data.Wrd[J];
                        Data.Wrd[J] := Xchg;
                        end;
          Inc(I);  Dec(J);    {Pass over exchanged records}
          end;
      until I > J;
    if Lo < J then QuickSort(Lo, J, A_LT_B, Self, Index);
    if I < Hi then QuickSort(I, Hi, A_LT_B, Self, Index);
  end;

  function Compare_Int(A, B: word; Data: BigStringPtr): boolean; far;
  begin
    with Data^ do
      Compare_Int := Data.Int[A] < Data.Int[B];
  end;

  function Compare_Wrd(A, B: word; Data: BigStringPtr): boolean; far;
  begin
    with Data^ do
      Compare_Wrd := Data.Wrd[A] < Data.Wrd[B];
  end;

  function Compare_Lng(A, B: word; Data: BigStringPtr): boolean; far;
  begin
    with Data^ do
      Compare_Lng := Data.Lng[A] < Data.Lng[B];
  end;

function BigStrings.Sort: BigStringPtr;
{Returns an index vector}
var
  Typ:    StringTypes;
  Len:    word;
  Result: BigStringPtr;
const
  Comparers: array[_int.._lng] of Comparer =
             (Compare_Int, Compare_Wrd, Compare_Lng);
begin
  Typ := StringType;
  case Typ of
    _Int, _Wrd: Len := Data.Wrd[0];
    _Lng:       Len := Data.Lng[0];
    else        RunError(201);
    end;
  Result := GenerateIndexVector(Len);
  QuickSort(1, Len, Comparers[Typ], @ Self, Result);
  Sort := Result;
end;

function BigStrings.Copy: BigStringPtr; {Returns a COPY of Self}
var
  Result: BigStringPtr;
begin
  GetMem(Result, ObjSize);
  Move(Self, Result^, ObjSize);
  Copy := Result;
end;

function BigStrings.EqualTo(Other: BigStringPtr): boolean;
var
  Idx: word;
begin
  EqualTo := False; {Unless it passes ALL tests}
  if StringType = Other^.StringType then
    case DataSize of
      2:   for Idx := 0 to Data.Wrd[0] do if
             Data.Wrd[Idx] <> Other^.Data.Wrd[Idx] then EXIT;
      4:   for Idx := 0 to Data.Lng[0] do if
             Data.Lng[Idx] <> Other^.Data.Lng[Idx] then EXIT;
      else RunError(201);
      end;
  EqualTo := True;
end;

end.