unit Consts; {Constants}

{This unit contains the key GA parameters, to make it easier to tweak them}

interface

const
  {GA unit}
  MultipleMutationRate = 150; {Per thousand}

  ShiftRatio     = 50;
  TransposeRatio = 25;
  ReverseRatio   = 05;

  All_Ratios     = ShiftRatio + TransposeRatio + ReverseRatio;

  ExpoRandRatio  = 666; {per thousand}

  {Top-level, flalloc program}
  BreedFrom      = 005;
  BreedTo        = 050;
  MaxGenerations = 250;

implementation

{* The meanings of the above constants:

   GA Unit:

     When the GA breeds a new gene string from an old one, it always makes at
     least one mutation.  MultipleMutationRate controls the rate of *multiple*
     mutations.  The default setting of 150 corresponds to 15% of all new
     genes being generated by two or more.  Note that the distribution of
     mutations is exponential: if 85% of all mutations are single mutations,
     85% of all multiple mutations are double mutations, and so on.

     The ShiftRatio, TransposeRatio, and ReverseRatio control the *types* of
     mutations.  The likelihood that any given mutation will be one of these
     three types is given by the ratio of it's constant to All_Ratios.  (Thus,
     with the default settings of 50/25/50, simple shifts are 5/8ths of all
     mutations.)  The SHIFT mutation takes a substring and moves it within the
     gene.  The TRANSPOSE mutation swaps two (identical length) substrings.
     The REVERSE mutation flips a substring, front to back (eg, 1 2 3 becomes
     3 2 1).

     The lengths of mutated substrings are exponentially distributed,
     according to the value of the ExpoRandatio.  At 666, 66.6% of all
     mutations involve more than one element; 66.% of all those involve more
     than two elements; and so on.  Thus, the higher ExpoRandRatio is, the
     longer the mutated substrings will be.

   All these values reflect my best guess at plausible settings -- I didn't
   do any experiments, here.

   Top level, flalloc program:

      BreedFrom and BreedTo control the size of the gene pool.  The GA
      generates BreedTo radndom initial genes, then culls that gene pool down
      to the BreedFrom best genes, then generates BreedTo-BreedFrom new
      candidates from the BreedFrom best, if the top scorer is not "perfect"
      and the maximum number of generations has not been exceeded.  I suspect
      that the best values for these parameters will depend on the particular
      problem you're trying to solve: for the relatively modest problems I
      tested FLALLOC on (up to ca 870 files, fitting on 36 720K floppies), I
      found that the seemingly low 5 and 50 worked best:  A higher BreedFrom
      just seemed to mean that more candidates were being generated from low
      quality parents, while a higher BreedTo seemed to make each
      generation take longer to generate and evaluate without yielding a
      higher rate of per-generation improvement.

      MaxGenerations controls how many generations should be tried before
      giving up.  The only time this came into play in my testing was when
      all the files were 200K and up -- ie, many floppies could only hold one
      file, and few (if any) could hold more than two or three.
*}

end.