program flalloc; {FLoppy space ALLOCate -
                    assign files to floppies,
                    minimizing total number of floppies}
{Copyright (C) 1994, Jon Shemitz <jon@armory.com> }

uses DOS,
     BigString, Dir_Fns,
     Consts, GA;

type
  String10 = string[10];
  String80 = string[80];

function zFmt(N: LongInt; Width: word): string10;
var
  Result: string;
begin
  Str(N, Result);
  if Length(Result) < Width then
    Result := Copy('0000000000', 1, Width - Length(Result)) + Result;
  zFmt := Result;
end;

function TimeNow: string80;
var
  H, M, S, Hu: word;
begin
  GetTime(H, M, S, Hu);
  TimeNow := zFmt(H, 2) + ':' + zFmt(M, 2) + ':' + zFmt(S, 2) + '.' + zFmt(Hu,2);
end;

var
  MainPath, InitPath: string[128];
  MainMask, InitMask: string[12];
  ClusterSize, ClustersPerDisk, InitClusters, OptimalNumberOfFloppies: word;

procedure ReadParamLine; {No error checking at all!}
var
  DriveLetter: string[1];
  DriveNumber: integer;
  Reg:         Dos.Registers;
  TotalInitClusters: LongInt;
begin
  MainPath := ParamStr(1);
  MainMask := ParamStr(2);
  DriveLetter := ParamStr(3);
  if (MainPath = '') or (MainMask = '') or (DriveLetter = '') then RunError(201);
  DriveNumber := ord(UpCase(DriveLetter[1])) - ord('A') + 1;
  if (DriveNumber <> 1) and (DriveNumber <> 2) then RunError(201);
  Reg.AH := $36; {Get Disk Free Space}
  Reg.DL := DriveNumber;
  MsDos(Reg);
  ClusterSize     := word(Reg.AX) * word(Reg.CX);
  ClustersPerDisk := Reg.DX;
  InitPath := ParamStr(4);
  InitMask := ParamStr(5);
  InitClusters := ClustersPerDisk;
  if (InitPath <> '') then
    InitClusters := ClustersPerDisk -
                    ScanDir( Nil, Nil,
                             InitPath, InitMask, ReadOnly+Archive,
                             ClusterSize, ClustersPerDisk );
end;

var
  PDR:         ProblemDescriptionRecord;

procedure ShowOrdering( Names:    FileNamesPtr;
                        Sizes:    BigStringPtr;
                        Ordering: GenePtr;
                        Tag:      string );
var
  Partition:   BigStringPtr;
  Floppy, Idx,
  FileNumber,
  Clusters,
  Limit:       word;
  Name:        string[12];
begin
  WriteLn('The ', Tag, ' ordering needs ', Floppies(PDR, Sizes, Ordering), ' floppies.');
  WriteLn('Score is ', Score(PDR, Sizes, Ordering));
  Partition := GA.Partition(PDR, Sizes, Ordering);
  Name[0] := #12;
  Idx := 0;
  for Floppy := 1 to Partition^.GetLength do
    begin
    WriteLn('Floppy ', Floppy, ': ',
            Partition^.Data.Wrd[Floppy] - Idx, ' files.');
    Clusters := 0;
    for Idx := Idx + 1 to Partition^.Data.Wrd[Floppy] do
      begin
      FileNumber := Ordering^.Data.Wrd[Idx];
      Move(Names^.Name[FileNumber], Name[1], 12);
      Inc(Clusters, Sizes^.Data.Wrd[FileNumber]);
      WriteLn(^I, Name, ': ', Sizes^.Data.Wrd[FileNumber]);
      end;
    if Floppy = 1 then Limit := InitClusters else Limit := ClustersPerDisk;
    WriteLn(Clusters, ' clusters (',
                      Round((mul_div(Clusters, 1000, Limit)+5)/10),
                      '%) used on this floppy.');
    WriteLn;
    end;
  WriteLn( 'This ', Tag, ' ordering needs ', Floppies(PDR, Sizes, Ordering),
           ' floppies; optimal is ', PDR.Optimal);

  Dispose_BigString(Partition);
end;

procedure OpenLogFile;
var
  LogFileName: string[80];
  LogFilePtr:  text;
begin
  Assign(LogFilePtr, 'log-file.nam');
  {$I-}
  Reset(LogfilePtr);
  if IoResult = 0
    then begin
         ReadLn(LogFilePtr, LogFileName);
         Close(LogFilePtr);
         end
    else LogFileName := '';
  Assign(Output, LogFileName);
  Rewrite(Output);
end;

var
  StartTime:     string80;
  TotalClusters: LongInt;
  Names:         FileNamesPtr;
  Sizes:         BigStringPtr;
  Ordering:      GenePtr;

begin
  OpenLogFile;
  StartTime := TimeNow;
  WriteLn('Start time:'^i, StartTime);

  Randomize;
  ReadParamLine;

  WriteLn('MainPath = '^I, MainPath);
  WriteLn('InitPath = '^I, InitPath);
  WriteLn;
  WriteLn('MainMask = '^I, MainMask);
  WriteLn('InitMask = '^I, InitMask);
  WriteLn;
  WriteLn('ClusterSize     ='^I, ClusterSize);
  WriteLn('ClustersPerDisk ='^I, ClustersPerDisk);
  WriteLn('InitClusters    ='^I, InitClusters);
  WriteLn;

  TotalClusters := ScanDir( @ Names, @ Sizes,
                            MainPath, MainMask, ReadOnly+Archive,
                            ClusterSize, ClustersPerDisk);
  OptimalNumberOfFloppies := (TotalClusters + (ClustersPerDisk - InitClusters) +
                              ClustersPerDisk - 1) div ClustersPerDisk;
  SetPDR(PDR, InitClusters, ClustersPerDisk, OptimalNumberOfFloppies);

  WriteLn( TotalClusters, ' clusters in ', Sizes^.GetLength, ' files: ',
           OptimalNumberOfFloppies, ' floppies is optimal.');
  WriteLn;

  Ordering := Optimize( PDR, Sizes, BreedFrom, BreedTo, MaxGenerations,
                        Output, True );
  ShowOrdering(Names, Sizes, Ordering, 'optimized');
  Dispose_BigString(Ordering);

  Dispose_FileNames(Names); Dispose_BigString(Sizes);

  WriteLn;
  WriteLn('Start time:'^I, StartTime);
  WriteLn('Stop time:'^I,  TimeNow);

  Close(Output);
end.